#include "othello.h"

#ifndef KANEKO
#define KANEKO

#ifdef KANEKODEBUG
#include <iostream.h>
#include <assert.h>
#endif

static const int search_max=100000, search_min=-100000;
static int search_lev=0;	// 

typedef char cint_t;
typedef cint_t piece_t;
typedef int index_t;
typedef char * const board_t;

typedef index_t loc_f(index_t);
typedef loc_f * loc_fptr;
typedef bool check_f(index_t);
typedef check_f * check_fptr;

inline piece_t p(board_t board, index_t i)
{
#ifdef KANEKODEBUG
  assert((0 <= i )&&( i < 64 ));
#endif
  return board[i];
}

inline index_t b_index(int x,int y){ return y*8+x; }	// 0,0 - 7,7

inline index_t left(index_t i){ return i-1; } 		// i ϰϤ
inline index_t right(index_t i){ return i+1; }
inline index_t up(index_t i){ return i-8; }
inline index_t down(index_t i){ return i+8; }

inline bool left_edge(index_t i){ return i % 8 == 0; }
inline bool right_edge(index_t i){ return i % 8 == 7; }
inline bool up_out(index_t i){ return i < 0; }
inline bool down_out(index_t i){ return i >= 64; }

inline bool my_colp(board_t board, index_t i){ return p(board,i)==MY_COLOR; }
inline bool op_colp(board_t board, index_t i){ return p(board,i)==MY_COLOR; }
//inline bool op_colp(board_t board, index_t i){ return p(board,i)==OP_COLOR; }
inline bool spacep(board_t board, index_t i){ return p(board,i)==SPACE; }

static const index_t
ul_c=b_index(0,0), ur_c=b_index(7,0), dl_c=b_index(0,7), dr_c=b_index(7,7);

static int
evalPoint( // int rest,
	  int real_point, int mobility, int cost_of_put,
	  int potential_count, int solid_edge,
	  int corner_point, int guard_point, int soft_edge)
{
  static const int corner_penalty=2;
  static const int edge_penalty=1;
  static const int mobility_bonus=2;
  static const int number_bonus=4;

  // դϤȤƤŬ :)
  int sum_number =
    ( real_point + cost_of_put )* number_bonus
      + ( mobility * mobility_bonus )
	+ potential_count / (( search_lev > 1 ) ? 16 : 15);
  
  int solid_special= ( search_lev > 1 ) ? 15 : 17;
  
  int sum_edge = soft_edge * edge_penalty;
  int sum_corner =
    (( solid_edge + guard_point) * solid_special
     + corner_point ) * corner_penalty;
  

  return sum_number + sum_edge + sum_corner;
}  
#ifdef KANEKODEBUG
static char
piece2char(piece_t p)
{
  switch(p)
    {
    case MY_COLOR:	return 'O';
    case OP_COLOR:	return 'X';
    case SPACE:		return '_';
    default:		return '?';
    }
}
static void				// fot debug
print_board(board_t board)
{
  bool endgame=true;
  for(int y=0;y<8;y++)
    {
      for(int x=0;x<8;x++)
	cerr << piece2char(p(board,b_index(x,y))) <<" ";
      cerr << endl;
    }
}
#endif

static int dummy1,dummy2;
inline static int
getRealPoint(board_t board, int & my_c=dummy1, int & op_c=dummy2)
{
  // --- number of stones ---
  CountStones(board,&my_c,&op_c);
  return op_c - my_c;
}

inline static int		// node start node 
checkEdge(board_t board, index_t node, loc_fptr mover,  bool & over_run)
{
  int adder =0;
  piece_t start = p(board,node);
  
  if     ( start == MY_COLOR )	adder = -1;
  else if( start == OP_COLOR )	adder =  1;
  else
    {
      over_run = false;
      return 0;
    }
  
  int sum=0; 
  int i=1;
  do{
    sum += adder;
    node = (*mover)(node);
  }while(( ++i<8 )&&( start == p(board,node) ));
  
  if( i == 8 )	over_run = true;
  else		over_run = false;
  
  return sum;
}

inline static bool
cornerGuard_ul_p(board_t board, piece_t teban)
{
  return ( p(board, right(ul_c)) == teban )
    && ( p(board, down(ul_c)) == teban )
      && ( p(board, right( down(ul_c))) == teban );
}
inline static bool
cornerGuard_ur_p(board_t board, piece_t teban)
{
  return
    ( p(board, left(ur_c)) == teban )
      && ( p(board, down(ur_c)) == teban )
	&& ( p(board, left(down(ur_c))) == teban );
}
inline static bool
cornerGuard_dl_p(board_t board, piece_t teban)
{
  return ( p(board, right(dl_c)) == teban )
    && ( p(board, up(dl_c)) == teban )
      && ( p(board, right(up(dl_c))) == teban) ;
}
inline static bool
cornerGuard_dr_p(board_t board, piece_t teban)
{
  return ( p(board, left(dr_c)) == teban )
    && ( p(board, up(dr_c)) == teban )
      && ( p(board, left(up(dr_c))) == teban );
}

static int
getPoint(board_t board)
{
  // 餱˶ᤤۤ礭

  // (0) --- real_point --- 
  int my_c,op_c;
  int real_point = getRealPoint(board,my_c,op_c);
  int rest = 64 - op_c - my_c;
  if(rest == 0)
    return real_point;
  if(my_c == 0) return search_max;
  if(op_c == 0) return search_min;
  
  // (1) --- mobility, cost_of_put, potential_count ---
  int my_cost=search_max, op_cost=search_max;	 // κǾȤɾ
  int my_mov=0, op_mov=0;
  int potential_count=0;
  for(index_t i=0; i<64; i++)
    {
      int point;
      if((point = Put(board,i,MY_COLOR,NOT_CHANGE)) > 0 )
	{
	  my_mov ++;
	  if( point < my_cost ) my_cost=point;
	  potential_count -= point;
	}
      if((point = Put(board,i,OP_COLOR,NOT_CHANGE)) > 0 )
	{
	  op_mov ++;
	  if( point < op_cost ) op_cost=point;
	  potential_count += point;
	}
    }
  if( my_mov == 0 ) my_cost = 0;
  if( op_mov == 0 ) op_cost = 0;
  
  int mobility = my_mov - op_mov;
  int cost_of_put = op_cost - my_cost;
  
  // (2) --- solid_edge --- 
  int solid_edge = 0;
  bool over_run;
  // [1] (0,0):0 <-> (7,0):7
  solid_edge += checkEdge( board, b_index(0,0), right, over_run );
  if(! over_run)
    solid_edge += checkEdge( board, b_index(7,0), left, over_run );
  
  // [2] (0,0):0 <-> (0,7):56
  solid_edge += checkEdge( board, b_index(0,0), down, over_run );
  if(! over_run)
    solid_edge += checkEdge( board, b_index(0,7), up, over_run );
  
  // [3] (7,7):63 <-> (0,7):56
  solid_edge += checkEdge( board, b_index(7,7), left, over_run );
  if(! over_run)
    solid_edge += checkEdge( board, b_index(0,7), right, over_run );
  // [4] (7,7):63 <-> (7,0):7
  solid_edge += checkEdge( board, b_index(7,7), up, over_run );
  if(! over_run)
    solid_edge += checkEdge( board, b_index(7,0), down, over_run );

  // (3) --- corner_point, guard_point --- 
  int corner_point=0, guard_point=0;
  
  index_t p = ul_c;
  if( spacep( board, p ))
    {
      if( Put(board,p,MY_COLOR,NOT_CHANGE)>0 )	corner_point--;
      if( Put(board,p,OP_COLOR,NOT_CHANGE)>0 )	corner_point++;
      if( cornerGuard_ul_p( board, MY_COLOR ))		guard_point++;
      else if( cornerGuard_ul_p( board, OP_COLOR ))	guard_point--;
    }
  
  p = ur_c;
  if( spacep( board, p ))
    {
      if( Put(board,p,MY_COLOR,NOT_CHANGE)>0 )	corner_point--;
      if( Put(board,p,OP_COLOR,NOT_CHANGE)>0 )	corner_point++;
      if( cornerGuard_ur_p( board, MY_COLOR ))		guard_point++;
      else if( cornerGuard_ur_p( board, OP_COLOR ))	guard_point--;
    }

  p = dl_c;
  if( spacep( board, p ))
    {
      if( Put(board,p,MY_COLOR,NOT_CHANGE)>0 )	corner_point--;
      if( Put(board,p,OP_COLOR,NOT_CHANGE)>0 )	corner_point++;
      if( cornerGuard_dl_p( board, MY_COLOR ))		guard_point++;
      else if( cornerGuard_dl_p( board, OP_COLOR ))	guard_point--;
    }

  p = dr_c;
  if( spacep( board,p ))
    {
      if( Put(board,p,MY_COLOR,NOT_CHANGE)>0 )	corner_point--;
      if( Put(board,p,OP_COLOR,NOT_CHANGE)>0 )	corner_point++;
      if( cornerGuard_dr_p( board, MY_COLOR ))		guard_point++;
      else if( cornerGuard_dr_p( board, OP_COLOR ))	guard_point--;
    }

  // (4) --- edge ---
  int soft_edge=0;
  for(index_t i=0;i<8;i++)
    {
      if( my_colp(board,i) ) soft_edge++;
      if( op_colp(board,i) ) soft_edge--;
    }
  for(index_t i=0;i<57;i+=8)
    {
      if( my_colp(board,i) ) soft_edge++;
      if( op_colp(board,i) ) soft_edge--;
    }
  for(index_t i=7;i<64;i+=8)
    {
      if( my_colp(board,i) ) soft_edge++;
      if( op_colp(board,i) ) soft_edge--;
    }
  for(index_t i=56;i<64;i++)
    {
      if( my_colp(board,i) ) soft_edge++;
      if( op_colp(board,i) ) soft_edge--;
    }
  
  return evalPoint( // rest,
		   real_point, mobility, cost_of_put, potential_count,
		   solid_edge, corner_point, guard_point, soft_edge);
}

static int
search(board_t board, int lev, int min, int max, piece_t teban,
       bool passed=false, bool expanded=false)
{
  //cerr << "Lev :"<<lev<<", Max :"<<max<<", Min :"<<min<<endl;
  if( lev < 1 )
    return getPoint(board);

  int put_count=0;
  char candidate[60];
  int minap=search_max;
  for(int i=0;i<64;i++)		// ǾͤƬŬsort
    {
      int ap=Put(board,i,teban,NOT_CHANGE);
      if( ap > 0)
	if( ap > minap )
	  candidate[ put_count++ ] = i;
	else
	  {
	    minap=ap;
	    candidate[ put_count++ ] = candidate[ 0 ];
	    candidate[ 0 ] = i;
	  }
    }
  if(( put_count < 3 )&&( lev < 3 )&&( !expanded ))
    {
      lev+=2;			// ưŪ
      expanded=true;
    }
  char newboard[64];
  CopyBoard(board, newboard, NORM);
  for(int j=0;j<put_count;j++)
    {
#ifdef KANEKODEBUG
      assert(( 0 <= candidate[j] )&&( candidate[j] < 64 ));
#endif
      CopyBoard(board, newboard, NORM);
      Put(newboard,candidate[j],teban,CHANGE);
      if( teban == MY_COLOR )	// max player
	{
	  int current=search(newboard,lev-1,min,max,OP_COLOR,false,expanded);
	  if ( min < current) min=current;
	  else if ( max <= current ) return current;	// beta-cut
	}
      else			// min player
	{
	  int current=search(newboard,lev-1,min,max,MY_COLOR,false,expanded);
	  if( current <= min ) return current;		// alpha-cut
	  else if ( current < max ) max = current;
	}
    }
  if( put_count>0 )
    if( teban == MY_COLOR )
      return min;
    else
      return max;
  else
    if( passed )
      {
	// print_board(board);
	// cerr <<((teban == MY_COLOR) ? " me ":" he ")<< "passed.\n";
	//return getPoint(board); 
	return getRealPoint(board);
      }
    else if( teban == MY_COLOR )
      return search(board,lev-1,min,max,OP_COLOR,true,expanded);
    else
      return search(board,lev-1,min,max,MY_COLOR,true,expanded);
#ifdef KANEKODEBUG
  cerr << "hoge?\n"; abort();
#endif
  return -1;
}
      
extern "C" int
kaneko(board_t board, int status)
{
  if(passp(status)) return 0;
  if(initp(status)) return 0;
#ifdef KANEKODEBUG
  cerr << "S";
#endif
  // ʤĤĴ٤
    
  int rest = 0;
  char candidate[60];
  int minap=search_max; int put_count=0;
  for(index_t i=0;i<64;i++)
    if( spacep(board, i) )
      {
	rest++;
	int ap=Put(board,i,MY_COLOR,NOT_CHANGE);
	if( ap > 0)
	  if( ap > minap )
	    candidate[ put_count++ ] = i;
	  else
	    {
	      minap=ap;
	      candidate[ put_count++ ] = candidate[ 0 ];
	      candidate[ 0 ] = i;
	    }
      }

  int lev;
  if( rest-search_lev < 10) lev = rest;
  else if (rest-search_lev <14 ) lev = search_lev+3; 
  else if ( rest-search_lev <16 ) lev = search_lev+2;
  else lev = search_lev;

  int loc=-1, point=search_min;
  char newboard[64];
  CopyBoard(board, newboard, NORM);
  for(int j=0;j<put_count;j++)
    {
      Put(newboard,candidate[j],MY_COLOR,CHANGE);
      int current = search( newboard, lev, point, search_max, OP_COLOR );
      if( current > point )
	{
	  point=current;
	  loc = candidate[j];
	}
      
      CopyBoard(board, newboard, NORM);
    }
#ifdef KANEKODEBUG
  cerr <<"Rest "<< rest <<"("<<lev<<") :  "<< point ;
  cerr <<" N "<< getPoint(board);
  cerr <<endl<<" (P:"<<put_count<<")";
  for(int i=0;i<put_count;i++)
    cerr <<" "<<(int)candidate[i];
  cerr <<endl;
  print_board(board);

  assert( loc >= 0 ); 
#endif
  return loc;
}

extern "C" int
kaneko0(board_t board, int status)
{
  search_lev = 0;
  return kaneko(board,status);
}
extern "C" int
kaneko1(board_t board, int status)
{
  search_lev = 1;
  return kaneko(board,status);
}
extern "C" int
kaneko2(board_t board, int status)
{
  search_lev = 2;
  return kaneko(board,status);
}
extern "C" int
kaneko3(board_t board, int status)
{
  search_lev = 3;
  return kaneko(board,status);
}

#endif
